@@ -280,6 +280,131 @@ def extractFontData(fontfiles, out):
                 f.writeByte(nftr.glyphs[chr(i)].length)
 
 
+# Archives
+class NARC:
+    btaf = 0
+    btnf = 0
+    gmif = 0
+    files = []
+
+
+class NARCFile:
+    start = 0
+    size = 0
+    path = ""
+    name = ""
+    fullname = ""
+
+
+def readNARC(narcfile):
+    common.logDebug("Reading", narcfile)
+    narc = NARC()
+    narc.files = []
+    with common.Stream(narcfile, "rb") as f:
+        # Read BTAF
+        f.seek(16)
+        narc.btaf = f.tell()
+        check = f.readString(4)
+        if check != "BTAF":
+            common.logError("Encountered", check, "instead of BTAF")
+            return None
+        sectionsize = f.readUInt()
+        narc.btnf = narc.btaf + sectionsize
+        filenum = f.readUInt()
+        common.logDebug("filenum:", filenum)
+        for i in range(filenum):
+            subfile = NARCFile()
+            subfile.start = narc.gmif + 8 + f.readUInt()
+            subfile.size = narc.gmif + 8 + f.readUInt() - subfile.start
+            common.logDebug(vars(subfile))
+            narc.files.append(subfile)
+        # Read BTNF
+        f.seek(narc.btnf)
+        check = f.readString(4)
+        if check != "BTNF":
+            common.logError("Encountered", check, "instead of BTNF")
+            return None
+        sectionsize = f.readUInt()
+        narc.gmif = narc.btnf + sectionsize
+        # TODO: handle directories
+        f.seek(8, 1)
+        for i in range(filenum):
+            namelen = f.readByte()
+            narc.files[i].name = f.readString(namelen)
+            narc.files[i].fullname = narc.files[i].path + narc.files[i].name
+        # Read GMIF
+        f.seek(narc.gmif)
+        check = f.readString(4)
+        if check != "GMIF":
+            common.logError("Encountered", check, "instead of GMIF")
+            return None
+    return narc
+
+
+def extractNARCFile(narcfile, outfolder):
+    narc = readNARC(narcfile)
+    if narc is None:
+        return
+    extractNARC(narcfile, outfolder, narc)
+
+
+def extractNARC(narcfile, outfolder, narc):
+    common.logDebug("Extracting", narcfile, "to", outfolder)
+    if not outfolder.endswith("/"):
+        outfolder = outfolder + "/"
+    common.makeFolder(outfolder)
+    with common.Stream(narcfile, "rb") as f:
+        for i in range(len(narc.files)):
+            file = narc.files[i]
+            f.seek(file.start)
+            with common.Stream(outfolder + file.fullname, "wb") as fout:
+                fout.write(f.read(file.size))
+
+
+def repackNARCFile(narcfilein, narcfileout, infolder):
+    narc = readNARC(narcfilein)
+    if narc is None:
+        return
+    repackNARC(narcfilein, narcfileout, infolder, narc)
+
+
+def repackNARC(narcfilein, narcfileout, infolder, narc):
+    common.logDebug("Repacking", narcfileout, "from", infolder)
+    with common.Stream(narcfilein, "rb") as fin:
+        with common.Stream(narcfileout, "wb") as f:
+            f.write(fin.read(narc.gmif + 8))
+            filepos = f.tell()
+            for i in range(len(narc.files)):
+                file = narc.files[i]
+                # Read file data
+                filepath = infolder + "/" + file.fullname
+                if not os.path.isfile(filepath):
+                    fin.seek(file.start)
+                    filedata = fin.read(file.size)
+                else:
+                    with common.Stream(filepath, "rb") as subf:
+                        filedata = subf.read()
+                # Write it in the archive
+                f.seek(filepos)
+                filestart = f.tell()
+                f.write(filedata)
+                fileend = f.tell()
+                # Pad with 0s
+                if f.tell() % 4 > 0:
+                    f.writeZero(f.tell() % 4)
+                filepos = f.tell()
+                # Update the pointers
+                f.seek(narc.btaf + 12 + i * 8)
+                f.writeUInt(filestart - narc.gmif - 8)
+                f.writeUInt(fileend - narc.gmif - 8)
+            # Write the new GMIF section size
+            f.seek(narc.gmif + 4)
+            f.writeUInt(filepos - narc.gmif)
+            # Write the new NARC size
+            f.seek(8)
+            f.writeUInt(filepos)
+
+
 # Graphics
 class NCGR:
     width = 0
@@ -485,6 +610,38 @@ def readMapData(data):
     return map
 
 
+def getNCERCellSize(shape, size):
+    cellsize = (0, 0)
+    if shape == 0:
+        if size == 0:
+            cellsize = (8, 8)
+        elif size == 1:
+            cellsize = (16, 16)
+        elif size == 2:
+            cellsize = (32, 32)
+        elif size == 3:
+            cellsize = (64, 64)
+    elif shape == 1:
+        if size == 0:
+            cellsize = (16, 8)
+        elif size == 1:
+            cellsize = (32, 8)
+        elif size == 2:
+            cellsize = (32, 16)
+        elif size == 3:
+            cellsize = (64, 32)
+    elif shape == 2:
+        if size == 0:
+            cellsize = (8, 16)
+        elif size == 1:
+            cellsize = (8, 32)
+        elif size == 2:
+            cellsize = (16, 32)
+        elif size == 3:
+            cellsize = (32, 64)
+    return cellsize
+
+
 def readNCER(ncerfile):
     ncer = NCER()
     ncer.banks = []
@@ -556,33 +713,7 @@ def readNCER(ncerfile):
                     cell.selectparam = (obj1 >> 9) & 0x1F
                 cell.priority = (obj2 >> 10) & 3
                 cell.pal = (obj2 >> 12) & 0xF
-                if cell.shape == 0:
-                    if cell.size == 0:
-                        cellsize = (8, 8)
-                    elif cell.size == 1:
-                        cellsize = (16, 16)
-                    elif cell.size == 2:
-                        cellsize = (32, 32)
-                    elif cell.size == 3:
-                        cellsize = (64, 64)
-                elif cell.shape == 1:
-                    if cell.size == 0:
-                        cellsize = (16, 8)
-                    elif cell.size == 1:
-                        cellsize = (32, 8)
-                    elif cell.size == 2:
-                        cellsize = (32, 16)
-                    elif cell.size == 3:
-                        cellsize = (64, 32)
-                elif cell.shape == 2:
-                    if cell.size == 0:
-                        cellsize = (8, 16)
-                    elif cell.size == 1:
-                        cellsize = (8, 32)
-                    elif cell.size == 2:
-                        cellsize = (16, 32)
-                    elif cell.size == 3:
-                        cellsize = (32, 64)
+                cellsize = getNCERCellSize(cell.shape, cell.size)
                 cell.width = cellsize[0]
                 cell.height = cellsize[1]
                 cell.numcell = j
@@ -861,7 +992,7 @@ def writeNSCR(file, ncgr, nscr, infile, palettes, width=-1, height=-1):
                 x += 1
 
 
-def writeMappedNSCR(file, mapfile, ncgr, nscr, infile, palettes, width=-1, height=-1, trasnptile=False, writelen=True):
+def writeMappedNSCR(file, mapfile, ncgr, nscr, infile, palettes, width=-1, height=-1, transptile=False, writelen=True):
     if width < 0:
         width = nscr.width
         # height = nscr.height
@@ -872,7 +1003,7 @@ def writeMappedNSCR(file, mapfile, ncgr, nscr, infile, palettes, width=-1, heigh
         with common.Stream(mapfile, "rb+") as mapf:
             mapf.seek(nscr.mapoffset)
             tiles = []
-            if trasnptile:
+            if transptile:
                 # Start with a completely transparent tile
                 tile = []
                 for i2 in range(ncgr.tilesize):
@@ -915,10 +1046,10 @@ def writeMappedNSCR(file, mapfile, ncgr, nscr, infile, palettes, width=-1, heigh
                 f.writeUInt(len(tiles) * (8 * ncgr.bpp))
 
 
-def writeMultiMappedNSCR(file, mapfiles, ncgr, nscrs, infiles, palettes, width=-1, height=-1, trasnptile=False, writelen=True):
+def writeMultiMappedNSCR(file, mapfiles, ncgr, nscrs, infiles, palettes, width=-1, height=-1, transptile=False, writelen=True):
     with common.Stream(file, "rb+") as f:
         tiles = []
-        if trasnptile:
+        if transptile:
             # Start with a completely transparent tile
             tile = []
             for i2 in range(ncgr.tilesize):
@@ -973,7 +1104,7 @@ def writeMultiMappedNSCR(file, mapfiles, ncgr, nscrs, infiles, palettes, width=-
             f.writeUInt(len(tiles) * (8 * ncgr.bpp))
 
 
-def writeNCER(file, ncerfile, ncgr, ncer, infile, palettes, width, height, appendTiles=False):
+def writeNCER(file, ncerfile, ncgr, ncer, infile, palettes, width=0, height=0, appendTiles=False, checkRepeat=True, writelen=True, fixtransp=False, checkalpha=False, zerotransp=True):
     psd = infile.endswith(".psd")
     if psd:
         psd = PSDImage.open(infile)
@@ -1024,7 +1155,7 @@ def writeNCER(file, ncerfile, ncgr, ncer, infile, palettes, width, height, appen
                     else:
                         pali = cell.pal * 16
                         palette = palettes[0]
-                    sametile = tile in donetiles
+                    sametile = checkRepeat and tile in donetiles
                     addingtiles = False
                     if sametile and appendTiles:
                         tiledata = []
@@ -1039,8 +1170,8 @@ def writeNCER(file, ncerfile, ncgr, ncer, infile, palettes, width, height, appen
                                         else:
                                             pixelx = cell.x + j * ncgr.tilesize + j2
                                             pixely = currheight + cell.y + i * ncgr.tilesize + i2
-                                        index1 = common.getPaletteIndex(palette, pixels[pixelx, pixely], False, pali, 16 if ncgr.bpp == 4 else -1)
-                                        index2 = common.getPaletteIndex(palette, pixels[pixelx + 1, pixely], False, pali, 16 if ncgr.bpp == 4 else -1)
+                                        index1 = common.getPaletteIndex(palette, pixels[pixelx, pixely], fixtransp, pali, 16 if ncgr.bpp == 4 else -1, checkalpha, zerotransp)
+                                        index2 = common.getPaletteIndex(palette, pixels[pixelx + 1, pixely], fixtransp, pali, 16 if ncgr.bpp == 4 else -1, checkalpha, zerotransp)
                                         tiledata.append(index1)
                                         tiledata.append(index2)
                         sametile = tiledata == cellboxes[tile]
@@ -1078,8 +1209,8 @@ def writeNCER(file, ncerfile, ncgr, ncer, infile, palettes, width, height, appen
                                             else:
                                                 pixelx = cell.x + j * ncgr.tilesize + j2
                                                 pixely = currheight + cell.y + i * ncgr.tilesize + i2
-                                            index1 = common.getPaletteIndex(palette, pixels[pixelx, pixely], False, pali, 16 if ncgr.bpp == 4 else -1)
-                                            index2 = common.getPaletteIndex(palette, pixels[pixelx + 1, pixely], False, pali, 16 if ncgr.bpp == 4 else -1)
+                                            index1 = common.getPaletteIndex(palette, pixels[pixelx, pixely], fixtransp, pali, 16 if ncgr.bpp == 4 else -1, checkalpha, zerotransp)
+                                            index2 = common.getPaletteIndex(palette, pixels[pixelx + 1, pixely], fixtransp, pali, 16 if ncgr.bpp == 4 else -1, checkalpha, zerotransp)
                                             cellboxes[currtile].append(index1)
                                             cellboxes[currtile].append(index2)
                                             writeNCGRData(f, ncgr.bpp, index1, index2)
@@ -1087,7 +1218,7 @@ def writeNCER(file, ncerfile, ncgr, ncer, infile, palettes, width, height, appen
                                 if addingtiles:
                                     nexttile += 1
                 currheight += bank.height
-        if nexttile > len(ncgr.tiles):
+        if writelen and nexttile > len(ncgr.tiles):
             tottiles = nexttile
             f.seek(32)
             f.writeUInt(tottiles)
